/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 18/04/22.
//

#include "fggl/gfx/ogl4/models.hpp"
#include "fggl/data/heightmap.hpp"

#include "fggl/gfx/camera.hpp"
#include "fggl/gfx/phong.hpp"

#include <spdlog/spdlog.h>

namespace fggl::gfx::ogl4 {

	static std::shared_ptr<ogl::ArrayBuffer> setupArrayBuffer(std::shared_ptr<ogl::VertexArray>& vao,
								 std::vector<data::Vertex>& data) {
		auto buff = std::make_shared<ogl::ArrayBuffer>();
		buff->write(data.size() * sizeof(data::Vertex), data.data(), ogl::BufUsage::STATIC_DRAW);

		// set up the vertex attributes
		auto posAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, posititon));
		auto normalAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, normal));
		auto colAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, colour));

		vao->setAttribute( *buff.get(), 0, posAttr );
		vao->setAttribute( *buff.get(), 1, normalAttr );
		vao->setAttribute( *buff.get(), 3, colAttr );
		return buff;
	}

	static std::shared_ptr<ogl::ElementBuffer> setupIndexBuffer(std::shared_ptr<ogl::VertexArray>& vao,
																std::vector<uint32_t>& data) {
		auto elementBuffer = std::make_shared<ogl::ElementBuffer>();
		elementBuffer->write(data.size() * sizeof(uint32_t),
							 data.data(), ogl::BufUsage::STATIC_DRAW);
		return elementBuffer;
	}

	static void setupComponent(StaticModel& modelComp, std::shared_ptr<ogl::Shader>& shader, data::Mesh& mesh) {
		auto vao = std::make_shared< ogl::VertexArray >();
		auto meshBuffer = setupArrayBuffer(vao, mesh.vertexList());
		auto elementBuffer = setupIndexBuffer(vao, mesh.indexList());

		// set up the element attributes
		modelComp.vao = vao;
		modelComp.vertexData = meshBuffer;
		modelComp.elements = elementBuffer;
		modelComp.pipeline = shader;
		modelComp.elementCount = mesh.indexCount();
		modelComp.drawType = ogl::Primative::TRIANGLE;
	}

	void StaticModelRenderer::resolveModels(entity::EntityManager &world) {
		// FIXME: this needs something reactive or performance will suck.
		auto renderables = world.find<data::StaticMesh>();
		for (const auto& renderable : renderables){
			auto* currModel = world.tryGet<StaticModel>( renderable );
			if ( currModel != nullptr ){
				continue;
			}

			auto& meshComp = world.get<data::StaticMesh>(renderable);
			auto& modelComp = world.add<StaticModel>(renderable);

			auto shader = m_phong;
			try {
				shader = std::make_shared<ogl::Shader>( m_shaders->get( meshComp.pipeline ) );
			} catch ( std::out_of_range& e) {
				debug::log(debug::Level::warning, "Could not find shader: {}", meshComp.pipeline);
			}

			setupComponent(modelComp, shader, meshComp.mesh);

			// no active model, we need to resolve/load one.
			//spdlog::info("looks like {} needs a static mesh", renderable);
		}

		// terrain
		auto terrain = world.find<data::HeightMap>();
		for (auto& renderable : terrain){
			auto currModel = world.tryGet<StaticModel>( renderable );
			if ( currModel != nullptr ){
				continue;
			}

			auto& heightmap = world.get<data::HeightMap>(renderable);
			data::Mesh heightMapMesh{};
			data::generateHeightMesh(heightmap, heightMapMesh);

			auto& modelComp = world.add<StaticModel>(renderable);
			setupComponent(modelComp, m_phong, heightMapMesh);

			// we know this is a triangle strip with a restart vertex...
			// FIXME the model should be telling us this...
			modelComp.drawType = ogl::Primative::TRIANGLE_STRIP;
			modelComp.restartIndex = heightMapMesh.restartVertex;

			// no active model, we need to resolve/load one.
			//spdlog::info("looks like heightmap, {} needs a static mesh", renderable);
		}
	}

	void StaticModelRenderer::renderModelsForward(const entity::EntityManager &world) {

		// fetch cameras we will need to render with
		auto cameras = world.find<gfx::Camera>();

		// if there are no cameras, we can't do anything...
		if ( cameras.size() == 0) {
			spdlog::warn("asked to render static models, but there were no cameras");
			return;
		}

		// perform a rendering pass for each camera (will usually only be one...)
		for ( auto& cameraEnt : cameras ){
			//TODO should be cliping this to only visible objects

			// enable required OpenGL state
			glEnable( GL_CULL_FACE );
			glCullFace( GL_BACK );

			// enable depth testing
			glEnable( GL_DEPTH_TEST );

			// set-up camera matrices
			const auto& camTransform = world.get<math::Transform>(cameraEnt);
			const auto& camComp = world.get<gfx::Camera>(cameraEnt);

			const math::mat4 projectionMatrix = glm::perspective(camComp.fov, camComp.aspectRatio, camComp.nearPlane, camComp.farPlane);
			const math::mat4 viewMatrix = glm::lookAt( camTransform.origin(), camComp.target, camTransform.up() );

			// TODO lighting needs to not be this...
			math::vec3 lightPos{0.0f, 10.0f, 0.0f};
			std::shared_ptr<ogl::Shader> shader = nullptr;

			auto renderables = world.find<StaticModel>();
			for ( const auto& entity : renderables ){
				// ensure that the model pipeline actually exists...
				const StaticModel& model = world.get<StaticModel>(entity);
				if ( model.pipeline == nullptr ) {
					spdlog::warn("shader was null, aborting render");
					continue;
				}

				// check if we switched shaders
				if ( shader != model.pipeline ) {
					// new shader - need to re-send the view and projection matrices
					shader = model.pipeline;
					shader->use();
					if ( shader->hasUniform("projection") ) {
						shader->setUniformMtx(shader->uniform("view"), viewMatrix);
						shader->setUniformMtx(shader->uniform("projection"), projectionMatrix);
					}
				}

				// set model transform
				const auto& transform = world.get<math::Transform>(entity);
				if (shader->hasUniform("model")) {
					shader->setUniformMtx(shader->uniform("model"), transform.model());
				} else {
					shader->setUniformMtx(shader->uniform("MVPMatrix"), projectionMatrix * viewMatrix * transform.model());
					shader->setUniformMtx(shader->uniform("MVMatrix"), viewMatrix * transform.model());

					auto normalMatrix = glm::mat3(glm::transpose(inverse(transform.model())));
					shader->setUniformMtx(shader->uniform("NormalMatrix"), normalMatrix);
				}

				// setup lighting mode
				if ( shader->hasUniform("lights[0].isEnabled") ) {
					bool local = true;

					shader->setUniformI(shader->uniform("lights[0].isEnabled"), 1);
					shader->setUniformI(shader->uniform("lights[0].isLocal"), local);
					shader->setUniformI(shader->uniform("lights[0].isSpot"), 0);

					shader->setUniformF(shader->uniform("lights[0].constantAttenuation"), 5.0f);
					shader->setUniformF(shader->uniform("lights[0].linearAttenuation"), 0.0f);
					shader->setUniformF(shader->uniform("lights[0].quadraticAttenuation"), 0.0f);

					shader->setUniformF(shader->uniform("Strength"), 0.6F);

					if (!local) {
						lightPos = glm::normalize(lightPos);
						auto viewDir = glm::normalize(camTransform.origin() - transform.origin());
						auto halfVector = glm::normalize(lightPos + viewDir);
						shader->setUniformF(shader->uniform("lights[0].halfVector"), halfVector );
						shader->setUniformF(shader->uniform("EyeDirection"), viewDir);
						shader->setUniformF(shader->uniform("lights[0].position"), lightPos);
					} else {
						auto camModelView = (viewMatrix * camTransform.model() * math::vec4(0.0f, 0.0f, 0.0f, 1.0f));
						auto modelModelView = (viewMatrix * transform.model() * math::vec4(0.0f, 0.0f, 0.0f, 1.0f));
						math::vec3 viewDir = glm::normalize(camModelView - modelModelView);
						shader->setUniformF(shader->uniform("EyeDirection"), viewDir);

						shader->setUniformF(shader->uniform("lights[0].position"), math::vec3(viewMatrix * math::vec4(lightPos, 1.0f)));
					}

					shader->setUniformF(shader->uniform("lights[0].ambient"), {0.0f, 0.5f, 0.0f});
					shader->setUniformF(shader->uniform("lights[0].colour"), {0.5f, 0.5f, 0.5f});
				}

				// material detection with fallback
				const auto& material = world.get<PhongMaterial>(entity);

				if ( shader->hasUniform("material.ambient") ) {
					shader->setUniformF(shader->uniform("material.ambient"), material.ambient);
					shader->setUniformF(shader->uniform("material.diffuse"), material.diffuse);
					shader->setUniformF(shader->uniform("material.specular"), material.specular);
					shader->setUniformF(shader->uniform("material.shininess"), material.shininess);
				} else {
					shader->setUniformF(shader->uniform("materials[0].emission"), material.emission);
					shader->setUniformF(shader->uniform("materials[0].ambient"), material.ambient);
					shader->setUniformF(shader->uniform("materials[0].diffuse"), material.diffuse);
					shader->setUniformF(shader->uniform("materials[0].specular"), material.specular);
					shader->setUniformF(shader->uniform("materials[0].shininess"), material.shininess);
				}

				if ( shader->hasUniform("lightPos")) {
					shader->setUniformF( shader->uniform("lightPos"), lightPos);
				}

				auto vao = model.vao;
				vao->bind();

				model.vertexData->bind();

				if ( model.restartIndex != NO_RESTART_IDX) {
					glEnable(GL_PRIMITIVE_RESTART);
					glPrimitiveRestartIndex(model.restartIndex);
				}
				auto elements = model.elements.get();
				vao->drawElements( *elements, model.drawType, model.elementCount);
				if ( model.restartIndex != NO_RESTART_IDX) {
					glDisable(GL_PRIMITIVE_RESTART);
				}
			}

		}
	}

}
