/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 24/07/22.
//

#ifndef FGGL_ENTITY_ENTITY_HPP
#define FGGL_ENTITY_ENTITY_HPP

#include <cstdint>
#include "fggl/vendor/entt.hpp"

namespace fggl::entity {

	using EntityID = entt::entity;
	constexpr EntityID INVALID = entt::null;

	class EntityManager {
		public:
			inline EntityID create() {
				return m_registry.create();
			}
			inline void destroy(EntityID entity) {
				m_registry.destroy(entity);
			}

			template<typename Component, typename... Args>
			inline Component& add(EntityID entity, Args&&... args) {
				return m_registry.emplace<Component>(entity, std::forward<Args>(args)...);
			}

			template<typename Component>
			Component& get(EntityID entity) {
				return m_registry.get<Component>(entity);
			}

			template<typename Component>
			const Component& get(EntityID entity) const {
				return m_registry.get<Component>(entity);
			}
			template<typename Component>
			Component* tryGet(EntityID entity) {
				return m_registry.try_get<Component>(entity);
			}

			template<typename Component>
			const Component* tryGet(EntityID entity) const {
				return m_registry.try_get<Component>(entity);
			}

			template<typename ...Components>
			auto find() const {
				return m_registry.view<Components...>();
			}

			template<typename ...Components>
			bool has(EntityID idx) const {
				return m_registry.template all_of<Components...>(idx);
			}

			inline bool exists(EntityID idx) {
				return m_registry.valid(idx);
			}

			inline bool alive(EntityID idx) {
				return m_registry.valid(idx);
			}


		private:
			entt::registry m_registry;
	};

	struct Entity {
		static Entity make(EntityManager& manager, EntityID idx) {
			return Entity{idx, manager};
		}

		EntityID id;
		EntityManager& manager;

		template<typename Component>
		Component& get() {
			return manager.get<Component>(id);
		}

		template<typename Component>
		const Component& get() const {
			return manager.get<Component>(id);
		}
	};

} // namespace fggl::entity

#endif //FGGL_ENTITY_ENTITY_HPP
